# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from collections import OrderedDict
import os
import re
from typing import (
    Dict,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from google.cloud.aiplatform_v1 import gapic_version as package_version

from google.api_core import client_options as client_options_lib
from google.api_core import exceptions as core_exceptions
from google.api_core import gapic_v1
from google.api_core import retry as retries
from google.auth import credentials as ga_credentials  # type: ignore
from google.auth.transport import mtls  # type: ignore
from google.auth.transport.grpc import SslCredentials  # type: ignore
from google.auth.exceptions import MutualTLSChannelError  # type: ignore
from google.oauth2 import service_account  # type: ignore

try:
    OptionalRetry = Union[retries.Retry, gapic_v1.method._MethodDefault]
except AttributeError:  # pragma: NO COVER
    OptionalRetry = Union[retries.Retry, object]  # type: ignore

from google.api import httpbody_pb2  # type: ignore
from google.cloud.aiplatform_v1.types import explanation
from google.cloud.aiplatform_v1.types import prediction_service
from google.cloud.location import locations_pb2  # type: ignore
from google.iam.v1 import iam_policy_pb2  # type: ignore
from google.iam.v1 import policy_pb2  # type: ignore
from google.longrunning import operations_pb2
from google.protobuf import any_pb2  # type: ignore
from google.protobuf import struct_pb2  # type: ignore
from .transports.base import PredictionServiceTransport, DEFAULT_CLIENT_INFO
from .transports.grpc import PredictionServiceGrpcTransport
from .transports.grpc_asyncio import PredictionServiceGrpcAsyncIOTransport


class PredictionServiceClientMeta(type):
    """Metaclass for the PredictionService client.

    This provides class-level methods for building and retrieving
    support objects (e.g. transport) without polluting the client instance
    objects.
    """

    _transport_registry = (
        OrderedDict()
    )  # type: Dict[str, Type[PredictionServiceTransport]]
    _transport_registry["grpc"] = PredictionServiceGrpcTransport
    _transport_registry["grpc_asyncio"] = PredictionServiceGrpcAsyncIOTransport

    def get_transport_class(
        cls,
        label: Optional[str] = None,
    ) -> Type[PredictionServiceTransport]:
        """Returns an appropriate transport class.

        Args:
            label: The name of the desired transport. If none is
                provided, then the first transport in the registry is used.

        Returns:
            The transport class to use.
        """
        # If a specific transport is requested, return that one.
        if label:
            return cls._transport_registry[label]

        # No transport is requested; return the default (that is, the first one
        # in the dictionary).
        return next(iter(cls._transport_registry.values()))


class PredictionServiceClient(metaclass=PredictionServiceClientMeta):
    """A service for online predictions and explanations."""

    @staticmethod
    def _get_default_mtls_endpoint(api_endpoint):
        """Converts api endpoint to mTLS endpoint.

        Convert "*.sandbox.googleapis.com" and "*.googleapis.com" to
        "*.mtls.sandbox.googleapis.com" and "*.mtls.googleapis.com" respectively.
        Args:
            api_endpoint (Optional[str]): the api endpoint to convert.
        Returns:
            str: converted mTLS api endpoint.
        """
        if not api_endpoint:
            return api_endpoint

        mtls_endpoint_re = re.compile(
            r"(?P<name>[^.]+)(?P<mtls>\.mtls)?(?P<sandbox>\.sandbox)?(?P<googledomain>\.googleapis\.com)?"
        )

        m = mtls_endpoint_re.match(api_endpoint)
        name, mtls, sandbox, googledomain = m.groups()
        if mtls or not googledomain:
            return api_endpoint

        if sandbox:
            return api_endpoint.replace(
                "sandbox.googleapis.com", "mtls.sandbox.googleapis.com"
            )

        return api_endpoint.replace(".googleapis.com", ".mtls.googleapis.com")

    DEFAULT_ENDPOINT = "aiplatform.googleapis.com"
    DEFAULT_MTLS_ENDPOINT = _get_default_mtls_endpoint.__func__(  # type: ignore
        DEFAULT_ENDPOINT
    )

    @classmethod
    def from_service_account_info(cls, info: dict, *args, **kwargs):
        """Creates an instance of this client using the provided credentials
            info.

        Args:
            info (dict): The service account private key info.
            args: Additional arguments to pass to the constructor.
            kwargs: Additional arguments to pass to the constructor.

        Returns:
            PredictionServiceClient: The constructed client.
        """
        credentials = service_account.Credentials.from_service_account_info(info)
        kwargs["credentials"] = credentials
        return cls(*args, **kwargs)

    @classmethod
    def from_service_account_file(cls, filename: str, *args, **kwargs):
        """Creates an instance of this client using the provided credentials
            file.

        Args:
            filename (str): The path to the service account private key json
                file.
            args: Additional arguments to pass to the constructor.
            kwargs: Additional arguments to pass to the constructor.

        Returns:
            PredictionServiceClient: The constructed client.
        """
        credentials = service_account.Credentials.from_service_account_file(filename)
        kwargs["credentials"] = credentials
        return cls(*args, **kwargs)

    from_service_account_json = from_service_account_file

    @property
    def transport(self) -> PredictionServiceTransport:
        """Returns the transport used by the client instance.

        Returns:
            PredictionServiceTransport: The transport used by the client
                instance.
        """
        return self._transport

    @staticmethod
    def endpoint_path(
        project: str,
        location: str,
        endpoint: str,
    ) -> str:
        """Returns a fully-qualified endpoint string."""
        return "projects/{project}/locations/{location}/endpoints/{endpoint}".format(
            project=project,
            location=location,
            endpoint=endpoint,
        )

    @staticmethod
    def parse_endpoint_path(path: str) -> Dict[str, str]:
        """Parses a endpoint path into its component segments."""
        m = re.match(
            r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/endpoints/(?P<endpoint>.+?)$",
            path,
        )
        return m.groupdict() if m else {}

    @staticmethod
    def model_path(
        project: str,
        location: str,
        model: str,
    ) -> str:
        """Returns a fully-qualified model string."""
        return "projects/{project}/locations/{location}/models/{model}".format(
            project=project,
            location=location,
            model=model,
        )

    @staticmethod
    def parse_model_path(path: str) -> Dict[str, str]:
        """Parses a model path into its component segments."""
        m = re.match(
            r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)/models/(?P<model>.+?)$",
            path,
        )
        return m.groupdict() if m else {}

    @staticmethod
    def common_billing_account_path(
        billing_account: str,
    ) -> str:
        """Returns a fully-qualified billing_account string."""
        return "billingAccounts/{billing_account}".format(
            billing_account=billing_account,
        )

    @staticmethod
    def parse_common_billing_account_path(path: str) -> Dict[str, str]:
        """Parse a billing_account path into its component segments."""
        m = re.match(r"^billingAccounts/(?P<billing_account>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_folder_path(
        folder: str,
    ) -> str:
        """Returns a fully-qualified folder string."""
        return "folders/{folder}".format(
            folder=folder,
        )

    @staticmethod
    def parse_common_folder_path(path: str) -> Dict[str, str]:
        """Parse a folder path into its component segments."""
        m = re.match(r"^folders/(?P<folder>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_organization_path(
        organization: str,
    ) -> str:
        """Returns a fully-qualified organization string."""
        return "organizations/{organization}".format(
            organization=organization,
        )

    @staticmethod
    def parse_common_organization_path(path: str) -> Dict[str, str]:
        """Parse a organization path into its component segments."""
        m = re.match(r"^organizations/(?P<organization>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_project_path(
        project: str,
    ) -> str:
        """Returns a fully-qualified project string."""
        return "projects/{project}".format(
            project=project,
        )

    @staticmethod
    def parse_common_project_path(path: str) -> Dict[str, str]:
        """Parse a project path into its component segments."""
        m = re.match(r"^projects/(?P<project>.+?)$", path)
        return m.groupdict() if m else {}

    @staticmethod
    def common_location_path(
        project: str,
        location: str,
    ) -> str:
        """Returns a fully-qualified location string."""
        return "projects/{project}/locations/{location}".format(
            project=project,
            location=location,
        )

    @staticmethod
    def parse_common_location_path(path: str) -> Dict[str, str]:
        """Parse a location path into its component segments."""
        m = re.match(r"^projects/(?P<project>.+?)/locations/(?P<location>.+?)$", path)
        return m.groupdict() if m else {}

    @classmethod
    def get_mtls_endpoint_and_cert_source(
        cls, client_options: Optional[client_options_lib.ClientOptions] = None
    ):
        """Return the API endpoint and client cert source for mutual TLS.

        The client cert source is determined in the following order:
        (1) if `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is not "true", the
        client cert source is None.
        (2) if `client_options.client_cert_source` is provided, use the provided one; if the
        default client cert source exists, use the default one; otherwise the client cert
        source is None.

        The API endpoint is determined in the following order:
        (1) if `client_options.api_endpoint` if provided, use the provided one.
        (2) if `GOOGLE_API_USE_CLIENT_CERTIFICATE` environment variable is "always", use the
        default mTLS endpoint; if the environment variable is "never", use the default API
        endpoint; otherwise if client cert source exists, use the default mTLS endpoint, otherwise
        use the default API endpoint.

        More details can be found at https://google.aip.dev/auth/4114.

        Args:
            client_options (google.api_core.client_options.ClientOptions): Custom options for the
                client. Only the `api_endpoint` and `client_cert_source` properties may be used
                in this method.

        Returns:
            Tuple[str, Callable[[], Tuple[bytes, bytes]]]: returns the API endpoint and the
                client cert source to use.

        Raises:
            google.auth.exceptions.MutualTLSChannelError: If any errors happen.
        """
        if client_options is None:
            client_options = client_options_lib.ClientOptions()
        use_client_cert = os.getenv("GOOGLE_API_USE_CLIENT_CERTIFICATE", "false")
        use_mtls_endpoint = os.getenv("GOOGLE_API_USE_MTLS_ENDPOINT", "auto")
        if use_client_cert not in ("true", "false"):
            raise ValueError(
                "Environment variable `GOOGLE_API_USE_CLIENT_CERTIFICATE` must be either `true` or `false`"
            )
        if use_mtls_endpoint not in ("auto", "never", "always"):
            raise MutualTLSChannelError(
                "Environment variable `GOOGLE_API_USE_MTLS_ENDPOINT` must be `never`, `auto` or `always`"
            )

        # Figure out the client cert source to use.
        client_cert_source = None
        if use_client_cert == "true":
            if client_options.client_cert_source:
                client_cert_source = client_options.client_cert_source
            elif mtls.has_default_client_cert_source():
                client_cert_source = mtls.default_client_cert_source()

        # Figure out which api endpoint to use.
        if client_options.api_endpoint is not None:
            api_endpoint = client_options.api_endpoint
        elif use_mtls_endpoint == "always" or (
            use_mtls_endpoint == "auto" and client_cert_source
        ):
            api_endpoint = cls.DEFAULT_MTLS_ENDPOINT
        else:
            api_endpoint = cls.DEFAULT_ENDPOINT

        return api_endpoint, client_cert_source

    def __init__(
        self,
        *,
        credentials: Optional[ga_credentials.Credentials] = None,
        transport: Optional[Union[str, PredictionServiceTransport]] = None,
        client_options: Optional[Union[client_options_lib.ClientOptions, dict]] = None,
        client_info: gapic_v1.client_info.ClientInfo = DEFAULT_CLIENT_INFO,
    ) -> None:
        """Instantiates the prediction service client.

        Args:
            credentials (Optional[google.auth.credentials.Credentials]): The
                authorization credentials to attach to requests. These
                credentials identify the application to the service; if none
                are specified, the client will attempt to ascertain the
                credentials from the environment.
            transport (Union[str, PredictionServiceTransport]): The
                transport to use. If set to None, a transport is chosen
                automatically.
            client_options (Optional[Union[google.api_core.client_options.ClientOptions, dict]]): Custom options for the
                client. It won't take effect if a ``transport`` instance is provided.
                (1) The ``api_endpoint`` property can be used to override the
                default endpoint provided by the client. GOOGLE_API_USE_MTLS_ENDPOINT
                environment variable can also be used to override the endpoint:
                "always" (always use the default mTLS endpoint), "never" (always
                use the default regular endpoint) and "auto" (auto switch to the
                default mTLS endpoint if client certificate is present, this is
                the default value). However, the ``api_endpoint`` property takes
                precedence if provided.
                (2) If GOOGLE_API_USE_CLIENT_CERTIFICATE environment variable
                is "true", then the ``client_cert_source`` property can be used
                to provide client certificate for mutual TLS transport. If
                not provided, the default SSL client certificate will be used if
                present. If GOOGLE_API_USE_CLIENT_CERTIFICATE is "false" or not
                set, no client certificate will be used.
            client_info (google.api_core.gapic_v1.client_info.ClientInfo):
                The client info used to send a user-agent string along with
                API requests. If ``None``, then default info will be used.
                Generally, you only need to set this if you're developing
                your own client library.

        Raises:
            google.auth.exceptions.MutualTLSChannelError: If mutual TLS transport
                creation failed for any reason.
        """
        if isinstance(client_options, dict):
            client_options = client_options_lib.from_dict(client_options)
        if client_options is None:
            client_options = client_options_lib.ClientOptions()
        client_options = cast(client_options_lib.ClientOptions, client_options)

        api_endpoint, client_cert_source_func = self.get_mtls_endpoint_and_cert_source(
            client_options
        )

        api_key_value = getattr(client_options, "api_key", None)
        if api_key_value and credentials:
            raise ValueError(
                "client_options.api_key and credentials are mutually exclusive"
            )

        # Save or instantiate the transport.
        # Ordinarily, we provide the transport, but allowing a custom transport
        # instance provides an extensibility point for unusual situations.
        if isinstance(transport, PredictionServiceTransport):
            # transport is a PredictionServiceTransport instance.
            if credentials or client_options.credentials_file or api_key_value:
                raise ValueError(
                    "When providing a transport instance, "
                    "provide its credentials directly."
                )
            if client_options.scopes:
                raise ValueError(
                    "When providing a transport instance, provide its scopes "
                    "directly."
                )
            self._transport = transport
        else:
            import google.auth._default  # type: ignore

            if api_key_value and hasattr(
                google.auth._default, "get_api_key_credentials"
            ):
                credentials = google.auth._default.get_api_key_credentials(
                    api_key_value
                )

            Transport = type(self).get_transport_class(transport)
            self._transport = Transport(
                credentials=credentials,
                credentials_file=client_options.credentials_file,
                host=api_endpoint,
                scopes=client_options.scopes,
                client_cert_source_for_mtls=client_cert_source_func,
                quota_project_id=client_options.quota_project_id,
                client_info=client_info,
                always_use_jwt_access=True,
                api_audience=client_options.api_audience,
            )

    def predict(
        self,
        request: Optional[Union[prediction_service.PredictRequest, dict]] = None,
        *,
        endpoint: Optional[str] = None,
        instances: Optional[MutableSequence[struct_pb2.Value]] = None,
        parameters: Optional[struct_pb2.Value] = None,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> prediction_service.PredictResponse:
        r"""Perform an online prediction.

        .. code-block:: python

            # This snippet has been automatically generated and should be regarded as a
            # code template only.
            # It will require modifications to work:
            # - It may require correct/in-range values for request initialization.
            # - It may require specifying regional endpoints when creating the service
            #   client as shown in:
            #   https://googleapis.dev/python/google-api-core/latest/client_options.html
            from google.cloud import aiplatform_v1

            def sample_predict():
                # Create a client
                client = aiplatform_v1.PredictionServiceClient()

                # Initialize request argument(s)
                instances = aiplatform_v1.Value()
                instances.null_value = "NULL_VALUE"

                request = aiplatform_v1.PredictRequest(
                    endpoint="endpoint_value",
                    instances=instances,
                )

                # Make the request
                response = client.predict(request=request)

                # Handle the response
                print(response)

        Args:
            request (Union[google.cloud.aiplatform_v1.types.PredictRequest, dict]):
                The request object. Request message for
                [PredictionService.Predict][google.cloud.aiplatform.v1.PredictionService.Predict].
            endpoint (str):
                Required. The name of the Endpoint requested to serve
                the prediction. Format:
                ``projects/{project}/locations/{location}/endpoints/{endpoint}``

                This corresponds to the ``endpoint`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            instances (MutableSequence[google.protobuf.struct_pb2.Value]):
                Required. The instances that are the input to the
                prediction call. A DeployedModel may have an upper limit
                on the number of instances it supports per request, and
                when it is exceeded the prediction call errors in case
                of AutoML Models, or, in case of customer created
                Models, the behaviour is as documented by that Model.
                The schema of any single instance may be specified via
                Endpoint's DeployedModels'
                [Model's][google.cloud.aiplatform.v1.DeployedModel.model]
                [PredictSchemata's][google.cloud.aiplatform.v1.Model.predict_schemata]
                [instance_schema_uri][google.cloud.aiplatform.v1.PredictSchemata.instance_schema_uri].

                This corresponds to the ``instances`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            parameters (google.protobuf.struct_pb2.Value):
                The parameters that govern the prediction. The schema of
                the parameters may be specified via Endpoint's
                DeployedModels' [Model's
                ][google.cloud.aiplatform.v1.DeployedModel.model]
                [PredictSchemata's][google.cloud.aiplatform.v1.Model.predict_schemata]
                [parameters_schema_uri][google.cloud.aiplatform.v1.PredictSchemata.parameters_schema_uri].

                This corresponds to the ``parameters`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            google.cloud.aiplatform_v1.types.PredictResponse:
                Response message for
                   [PredictionService.Predict][google.cloud.aiplatform.v1.PredictionService.Predict].

        """
        # Create or coerce a protobuf request object.
        # Quick check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([endpoint, instances, parameters])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a prediction_service.PredictRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, prediction_service.PredictRequest):
            request = prediction_service.PredictRequest(request)
            # If we have keyword arguments corresponding to fields on the
            # request, apply these.
            if endpoint is not None:
                request.endpoint = endpoint
            if instances is not None:
                request.instances.extend(instances)
            if parameters is not None:
                request.parameters = parameters

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.predict]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("endpoint", request.endpoint),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def raw_predict(
        self,
        request: Optional[Union[prediction_service.RawPredictRequest, dict]] = None,
        *,
        endpoint: Optional[str] = None,
        http_body: Optional[httpbody_pb2.HttpBody] = None,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> httpbody_pb2.HttpBody:
        r"""Perform an online prediction with an arbitrary HTTP payload.

        The response includes the following HTTP headers:

        -  ``X-Vertex-AI-Endpoint-Id``: ID of the
           [Endpoint][google.cloud.aiplatform.v1.Endpoint] that served
           this prediction.

        -  ``X-Vertex-AI-Deployed-Model-Id``: ID of the Endpoint's
           [DeployedModel][google.cloud.aiplatform.v1.DeployedModel]
           that served this prediction.

        .. code-block:: python

            # This snippet has been automatically generated and should be regarded as a
            # code template only.
            # It will require modifications to work:
            # - It may require correct/in-range values for request initialization.
            # - It may require specifying regional endpoints when creating the service
            #   client as shown in:
            #   https://googleapis.dev/python/google-api-core/latest/client_options.html
            from google.cloud import aiplatform_v1

            def sample_raw_predict():
                # Create a client
                client = aiplatform_v1.PredictionServiceClient()

                # Initialize request argument(s)
                request = aiplatform_v1.RawPredictRequest(
                    endpoint="endpoint_value",
                )

                # Make the request
                response = client.raw_predict(request=request)

                # Handle the response
                print(response)

        Args:
            request (Union[google.cloud.aiplatform_v1.types.RawPredictRequest, dict]):
                The request object. Request message for
                [PredictionService.RawPredict][google.cloud.aiplatform.v1.PredictionService.RawPredict].
            endpoint (str):
                Required. The name of the Endpoint requested to serve
                the prediction. Format:
                ``projects/{project}/locations/{location}/endpoints/{endpoint}``

                This corresponds to the ``endpoint`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            http_body (google.api.httpbody_pb2.HttpBody):
                The prediction input. Supports HTTP headers and
                arbitrary data payload.

                A
                [DeployedModel][google.cloud.aiplatform.v1.DeployedModel]
                may have an upper limit on the number of instances it
                supports per request. When this limit it is exceeded for
                an AutoML model, the
                [RawPredict][google.cloud.aiplatform.v1.PredictionService.RawPredict]
                method returns an error. When this limit is exceeded for
                a custom-trained model, the behavior varies depending on
                the model.

                You can specify the schema for each instance in the
                [predict_schemata.instance_schema_uri][google.cloud.aiplatform.v1.PredictSchemata.instance_schema_uri]
                field when you create a
                [Model][google.cloud.aiplatform.v1.Model]. This schema
                applies when you deploy the ``Model`` as a
                ``DeployedModel`` to an
                [Endpoint][google.cloud.aiplatform.v1.Endpoint] and use
                the ``RawPredict`` method.

                This corresponds to the ``http_body`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            google.api.httpbody_pb2.HttpBody:
                Message that represents an arbitrary HTTP body. It should only be used for
                   payload formats that can't be represented as JSON,
                   such as raw binary or an HTML page.

                   This message can be used both in streaming and
                   non-streaming API methods in the request as well as
                   the response.

                   It can be used as a top-level request field, which is
                   convenient if one wants to extract parameters from
                   either the URL or HTTP template into the request
                   fields and also want access to the raw HTTP body.

                   Example:

                      message GetResourceRequest {
                         // A unique request id. string request_id = 1;

                         // The raw HTTP body is bound to this field.
                         google.api.HttpBody http_body = 2;

                      }

                      service ResourceService {
                         rpc GetResource(GetResourceRequest)
                            returns (google.api.HttpBody);

                         rpc UpdateResource(google.api.HttpBody)
                            returns (google.protobuf.Empty);

                      }

                   Example with streaming methods:

                      service CaldavService {
                         rpc GetCalendar(stream google.api.HttpBody)
                            returns (stream google.api.HttpBody);

                         rpc UpdateCalendar(stream google.api.HttpBody)
                            returns (stream google.api.HttpBody);

                      }

                   Use of this type only changes how the request and
                   response bodies are handled, all other features will
                   continue to work unchanged.

        """
        # Create or coerce a protobuf request object.
        # Quick check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([endpoint, http_body])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a prediction_service.RawPredictRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, prediction_service.RawPredictRequest):
            request = prediction_service.RawPredictRequest(request)
            # If we have keyword arguments corresponding to fields on the
            # request, apply these.
            if endpoint is not None:
                request.endpoint = endpoint
            if http_body is not None:
                request.http_body = http_body

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.raw_predict]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("endpoint", request.endpoint),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def explain(
        self,
        request: Optional[Union[prediction_service.ExplainRequest, dict]] = None,
        *,
        endpoint: Optional[str] = None,
        instances: Optional[MutableSequence[struct_pb2.Value]] = None,
        parameters: Optional[struct_pb2.Value] = None,
        deployed_model_id: Optional[str] = None,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> prediction_service.ExplainResponse:
        r"""Perform an online explanation.

        If
        [deployed_model_id][google.cloud.aiplatform.v1.ExplainRequest.deployed_model_id]
        is specified, the corresponding DeployModel must have
        [explanation_spec][google.cloud.aiplatform.v1.DeployedModel.explanation_spec]
        populated. If
        [deployed_model_id][google.cloud.aiplatform.v1.ExplainRequest.deployed_model_id]
        is not specified, all DeployedModels must have
        [explanation_spec][google.cloud.aiplatform.v1.DeployedModel.explanation_spec]
        populated. Only deployed AutoML tabular Models have
        explanation_spec.

        .. code-block:: python

            # This snippet has been automatically generated and should be regarded as a
            # code template only.
            # It will require modifications to work:
            # - It may require correct/in-range values for request initialization.
            # - It may require specifying regional endpoints when creating the service
            #   client as shown in:
            #   https://googleapis.dev/python/google-api-core/latest/client_options.html
            from google.cloud import aiplatform_v1

            def sample_explain():
                # Create a client
                client = aiplatform_v1.PredictionServiceClient()

                # Initialize request argument(s)
                instances = aiplatform_v1.Value()
                instances.null_value = "NULL_VALUE"

                request = aiplatform_v1.ExplainRequest(
                    endpoint="endpoint_value",
                    instances=instances,
                )

                # Make the request
                response = client.explain(request=request)

                # Handle the response
                print(response)

        Args:
            request (Union[google.cloud.aiplatform_v1.types.ExplainRequest, dict]):
                The request object. Request message for
                [PredictionService.Explain][google.cloud.aiplatform.v1.PredictionService.Explain].
            endpoint (str):
                Required. The name of the Endpoint requested to serve
                the explanation. Format:
                ``projects/{project}/locations/{location}/endpoints/{endpoint}``

                This corresponds to the ``endpoint`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            instances (MutableSequence[google.protobuf.struct_pb2.Value]):
                Required. The instances that are the input to the
                explanation call. A DeployedModel may have an upper
                limit on the number of instances it supports per
                request, and when it is exceeded the explanation call
                errors in case of AutoML Models, or, in case of customer
                created Models, the behaviour is as documented by that
                Model. The schema of any single instance may be
                specified via Endpoint's DeployedModels'
                [Model's][google.cloud.aiplatform.v1.DeployedModel.model]
                [PredictSchemata's][google.cloud.aiplatform.v1.Model.predict_schemata]
                [instance_schema_uri][google.cloud.aiplatform.v1.PredictSchemata.instance_schema_uri].

                This corresponds to the ``instances`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            parameters (google.protobuf.struct_pb2.Value):
                The parameters that govern the prediction. The schema of
                the parameters may be specified via Endpoint's
                DeployedModels' [Model's
                ][google.cloud.aiplatform.v1.DeployedModel.model]
                [PredictSchemata's][google.cloud.aiplatform.v1.Model.predict_schemata]
                [parameters_schema_uri][google.cloud.aiplatform.v1.PredictSchemata.parameters_schema_uri].

                This corresponds to the ``parameters`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            deployed_model_id (str):
                If specified, this ExplainRequest will be served by the
                chosen DeployedModel, overriding
                [Endpoint.traffic_split][google.cloud.aiplatform.v1.Endpoint.traffic_split].

                This corresponds to the ``deployed_model_id`` field
                on the ``request`` instance; if ``request`` is provided, this
                should not be set.
            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.

        Returns:
            google.cloud.aiplatform_v1.types.ExplainResponse:
                Response message for
                   [PredictionService.Explain][google.cloud.aiplatform.v1.PredictionService.Explain].

        """
        # Create or coerce a protobuf request object.
        # Quick check: If we got a request object, we should *not* have
        # gotten any keyword arguments that map to the request.
        has_flattened_params = any([endpoint, instances, parameters, deployed_model_id])
        if request is not None and has_flattened_params:
            raise ValueError(
                "If the `request` argument is set, then none of "
                "the individual field arguments should be set."
            )

        # Minor optimization to avoid making a copy if the user passes
        # in a prediction_service.ExplainRequest.
        # There's no risk of modifying the input as we've already verified
        # there are no flattened fields.
        if not isinstance(request, prediction_service.ExplainRequest):
            request = prediction_service.ExplainRequest(request)
            # If we have keyword arguments corresponding to fields on the
            # request, apply these.
            if endpoint is not None:
                request.endpoint = endpoint
            if instances is not None:
                request.instances.extend(instances)
            if parameters is not None:
                request.parameters = parameters
            if deployed_model_id is not None:
                request.deployed_model_id = deployed_model_id

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = self._transport._wrapped_methods[self._transport.explain]

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("endpoint", request.endpoint),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def __enter__(self) -> "PredictionServiceClient":
        return self

    def __exit__(self, type, value, traceback):
        """Releases underlying transport's resources.

        .. warning::
            ONLY use as a context manager if the transport is NOT shared
            with other clients! Exiting the with block will CLOSE the transport
            and may cause errors in other clients!
        """
        self.transport.close()

    def list_operations(
        self,
        request: Optional[operations_pb2.ListOperationsRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> operations_pb2.ListOperationsResponse:
        r"""Lists operations that match the specified filter in the request.

        Args:
            request (:class:`~.operations_pb2.ListOperationsRequest`):
                The request object. Request message for
                `ListOperations` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                    if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.operations_pb2.ListOperationsResponse:
                Response message for ``ListOperations`` method.
        """
        # Create or coerce a protobuf request object.
        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = operations_pb2.ListOperationsRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.list_operations,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def get_operation(
        self,
        request: Optional[operations_pb2.GetOperationRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> operations_pb2.Operation:
        r"""Gets the latest state of a long-running operation.

        Args:
            request (:class:`~.operations_pb2.GetOperationRequest`):
                The request object. Request message for
                `GetOperation` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                    if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.operations_pb2.Operation:
                An ``Operation`` object.
        """
        # Create or coerce a protobuf request object.
        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = operations_pb2.GetOperationRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.get_operation,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def delete_operation(
        self,
        request: Optional[operations_pb2.DeleteOperationRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> None:
        r"""Deletes a long-running operation.

        This method indicates that the client is no longer interested
        in the operation result. It does not cancel the operation.
        If the server doesn't support this method, it returns
        `google.rpc.Code.UNIMPLEMENTED`.

        Args:
            request (:class:`~.operations_pb2.DeleteOperationRequest`):
                The request object. Request message for
                `DeleteOperation` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                    if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            None
        """
        # Create or coerce a protobuf request object.
        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = operations_pb2.DeleteOperationRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.delete_operation,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    def cancel_operation(
        self,
        request: Optional[operations_pb2.CancelOperationRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> None:
        r"""Starts asynchronous cancellation on a long-running operation.

        The server makes a best effort to cancel the operation, but success
        is not guaranteed.  If the server doesn't support this method, it returns
        `google.rpc.Code.UNIMPLEMENTED`.

        Args:
            request (:class:`~.operations_pb2.CancelOperationRequest`):
                The request object. Request message for
                `CancelOperation` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                    if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            None
        """
        # Create or coerce a protobuf request object.
        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = operations_pb2.CancelOperationRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.cancel_operation,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

    def wait_operation(
        self,
        request: Optional[operations_pb2.WaitOperationRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> operations_pb2.Operation:
        r"""Waits until the specified long-running operation is done or reaches at most
        a specified timeout, returning the latest state.

        If the operation is already done, the latest state is immediately returned.
        If the timeout specified is greater than the default HTTP/RPC timeout, the HTTP/RPC
        timeout is used.  If the server does not support this method, it returns
        `google.rpc.Code.UNIMPLEMENTED`.

        Args:
            request (:class:`~.operations_pb2.WaitOperationRequest`):
                The request object. Request message for
                `WaitOperation` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                    if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.operations_pb2.Operation:
                An ``Operation`` object.
        """
        # Create or coerce a protobuf request object.
        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = operations_pb2.WaitOperationRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.wait_operation,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def set_iam_policy(
        self,
        request: Optional[iam_policy_pb2.SetIamPolicyRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> policy_pb2.Policy:
        r"""Sets the IAM access control policy on the specified function.

        Replaces any existing policy.

        Args:
            request (:class:`~.iam_policy_pb2.SetIamPolicyRequest`):
                The request object. Request message for `SetIamPolicy`
                method.
            retry (google.api_core.retry.Retry): Designation of what errors, if any,
                should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.policy_pb2.Policy:
                Defines an Identity and Access Management (IAM) policy.
                It is used to specify access control policies for Cloud
                Platform resources.
                A ``Policy`` is a collection of ``bindings``. A
                ``binding`` binds one or more ``members`` to a single
                ``role``. Members can be user accounts, service
                accounts, Google groups, and domains (such as G Suite).
                A ``role`` is a named list of permissions (defined by
                IAM or configured by users). A ``binding`` can
                optionally specify a ``condition``, which is a logic
                expression that further constrains the role binding
                based on attributes about the request and/or target
                resource.

                **JSON Example**

                ::

                    {
                      "bindings": [
                        {
                          "role": "roles/resourcemanager.organizationAdmin",
                          "members": [
                            "user:mike@example.com",
                            "group:admins@example.com",
                            "domain:google.com",
                            "serviceAccount:my-project-id@appspot.gserviceaccount.com"
                          ]
                        },
                        {
                          "role": "roles/resourcemanager.organizationViewer",
                          "members": ["user:eve@example.com"],
                          "condition": {
                            "title": "expirable access",
                            "description": "Does not grant access after Sep 2020",
                            "expression": "request.time <
                            timestamp('2020-10-01T00:00:00.000Z')",
                          }
                        }
                      ]
                    }

                **YAML Example**

                ::

                    bindings:
                    - members:
                      - user:mike@example.com
                      - group:admins@example.com
                      - domain:google.com
                      - serviceAccount:my-project-id@appspot.gserviceaccount.com
                      role: roles/resourcemanager.organizationAdmin
                    - members:
                      - user:eve@example.com
                      role: roles/resourcemanager.organizationViewer
                      condition:
                        title: expirable access
                        description: Does not grant access after Sep 2020
                        expression: request.time < timestamp('2020-10-01T00:00:00.000Z')

                For a description of IAM and its features, see the `IAM
                developer's
                guide <https://cloud.google.com/iam/docs>`__.
        """
        # Create or coerce a protobuf request object.

        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = iam_policy_pb2.SetIamPolicyRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.set_iam_policy,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("resource", request.resource),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def get_iam_policy(
        self,
        request: Optional[iam_policy_pb2.GetIamPolicyRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> policy_pb2.Policy:
        r"""Gets the IAM access control policy for a function.

        Returns an empty policy if the function exists and does not have a
        policy set.

        Args:
            request (:class:`~.iam_policy_pb2.GetIamPolicyRequest`):
                The request object. Request message for `GetIamPolicy`
                method.
            retry (google.api_core.retry.Retry): Designation of what errors, if
                any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.policy_pb2.Policy:
                Defines an Identity and Access Management (IAM) policy.
                It is used to specify access control policies for Cloud
                Platform resources.
                A ``Policy`` is a collection of ``bindings``. A
                ``binding`` binds one or more ``members`` to a single
                ``role``. Members can be user accounts, service
                accounts, Google groups, and domains (such as G Suite).
                A ``role`` is a named list of permissions (defined by
                IAM or configured by users). A ``binding`` can
                optionally specify a ``condition``, which is a logic
                expression that further constrains the role binding
                based on attributes about the request and/or target
                resource.

                **JSON Example**

                ::

                    {
                      "bindings": [
                        {
                          "role": "roles/resourcemanager.organizationAdmin",
                          "members": [
                            "user:mike@example.com",
                            "group:admins@example.com",
                            "domain:google.com",
                            "serviceAccount:my-project-id@appspot.gserviceaccount.com"
                          ]
                        },
                        {
                          "role": "roles/resourcemanager.organizationViewer",
                          "members": ["user:eve@example.com"],
                          "condition": {
                            "title": "expirable access",
                            "description": "Does not grant access after Sep 2020",
                            "expression": "request.time <
                            timestamp('2020-10-01T00:00:00.000Z')",
                          }
                        }
                      ]
                    }

                **YAML Example**

                ::

                    bindings:
                    - members:
                      - user:mike@example.com
                      - group:admins@example.com
                      - domain:google.com
                      - serviceAccount:my-project-id@appspot.gserviceaccount.com
                      role: roles/resourcemanager.organizationAdmin
                    - members:
                      - user:eve@example.com
                      role: roles/resourcemanager.organizationViewer
                      condition:
                        title: expirable access
                        description: Does not grant access after Sep 2020
                        expression: request.time < timestamp('2020-10-01T00:00:00.000Z')

                For a description of IAM and its features, see the `IAM
                developer's
                guide <https://cloud.google.com/iam/docs>`__.
        """
        # Create or coerce a protobuf request object.

        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = iam_policy_pb2.GetIamPolicyRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.get_iam_policy,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("resource", request.resource),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def test_iam_permissions(
        self,
        request: Optional[iam_policy_pb2.TestIamPermissionsRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> iam_policy_pb2.TestIamPermissionsResponse:
        r"""Tests the specified IAM permissions against the IAM access control
            policy for a function.

        If the function does not exist, this will return an empty set
        of permissions, not a NOT_FOUND error.

        Args:
            request (:class:`~.iam_policy_pb2.TestIamPermissionsRequest`):
                The request object. Request message for
                `TestIamPermissions` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                 if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.iam_policy_pb2.TestIamPermissionsResponse:
                Response message for ``TestIamPermissions`` method.
        """
        # Create or coerce a protobuf request object.

        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = iam_policy_pb2.TestIamPermissionsRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.test_iam_permissions,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("resource", request.resource),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def get_location(
        self,
        request: Optional[locations_pb2.GetLocationRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> locations_pb2.Location:
        r"""Gets information about a location.

        Args:
            request (:class:`~.location_pb2.GetLocationRequest`):
                The request object. Request message for
                `GetLocation` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                 if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.location_pb2.Location:
                Location object.
        """
        # Create or coerce a protobuf request object.
        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = locations_pb2.GetLocationRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.get_location,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response

    def list_locations(
        self,
        request: Optional[locations_pb2.ListLocationsRequest] = None,
        *,
        retry: OptionalRetry = gapic_v1.method.DEFAULT,
        timeout: Union[float, object] = gapic_v1.method.DEFAULT,
        metadata: Sequence[Tuple[str, str]] = (),
    ) -> locations_pb2.ListLocationsResponse:
        r"""Lists information about the supported locations for this service.

        Args:
            request (:class:`~.location_pb2.ListLocationsRequest`):
                The request object. Request message for
                `ListLocations` method.
            retry (google.api_core.retry.Retry): Designation of what errors,
                 if any, should be retried.
            timeout (float): The timeout for this request.
            metadata (Sequence[Tuple[str, str]]): Strings which should be
                sent along with the request as metadata.
        Returns:
            ~.location_pb2.ListLocationsResponse:
                Response message for ``ListLocations`` method.
        """
        # Create or coerce a protobuf request object.
        # The request isn't a proto-plus wrapped type,
        # so it must be constructed via keyword expansion.
        if isinstance(request, dict):
            request = locations_pb2.ListLocationsRequest(**request)

        # Wrap the RPC method; this adds retry and timeout information,
        # and friendly error handling.
        rpc = gapic_v1.method.wrap_method(
            self._transport.list_locations,
            default_timeout=None,
            client_info=DEFAULT_CLIENT_INFO,
        )

        # Certain fields should be provided within the metadata header;
        # add these here.
        metadata = tuple(metadata) + (
            gapic_v1.routing_header.to_grpc_metadata((("name", request.name),)),
        )

        # Send the request.
        response = rpc(
            request,
            retry=retry,
            timeout=timeout,
            metadata=metadata,
        )

        # Done; return the response.
        return response


DEFAULT_CLIENT_INFO = gapic_v1.client_info.ClientInfo(
    gapic_version=package_version.__version__
)


__all__ = ("PredictionServiceClient",)



# -*- coding: utf-8 -*-
# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from google.cloud.aiplatform_v1beta1 import gapic_version as package_version

__version__ = package_version.__version__


from .services.dataset_service import DatasetServiceClient
from .services.dataset_service import DatasetServiceAsyncClient
from .services.deployment_resource_pool_service import (
    DeploymentResourcePoolServiceClient,
)
from .services.deployment_resource_pool_service import (
    DeploymentResourcePoolServiceAsyncClient,
)
from .services.endpoint_service import EndpointServiceClient
from .services.endpoint_service import EndpointServiceAsyncClient
from .services.featurestore_online_serving_service import (
    FeaturestoreOnlineServingServiceClient,
)
from .services.featurestore_online_serving_service import (
    FeaturestoreOnlineServingServiceAsyncClient,
)
from .services.featurestore_service import FeaturestoreServiceClient
from .services.featurestore_service import FeaturestoreServiceAsyncClient
from .services.index_endpoint_service import IndexEndpointServiceClient
from .services.index_endpoint_service import IndexEndpointServiceAsyncClient
from .services.index_service import IndexServiceClient
from .services.index_service import IndexServiceAsyncClient
from .services.job_service import JobServiceClient
from .services.job_service import JobServiceAsyncClient
from .services.metadata_service import MetadataServiceClient
from .services.metadata_service import MetadataServiceAsyncClient
from .services.migration_service import MigrationServiceClient
from .services.migration_service import MigrationServiceAsyncClient
from .services.model_service import ModelServiceClient
from .services.model_service import ModelServiceAsyncClient
from .services.pipeline_service import PipelineServiceClient
from .services.pipeline_service import PipelineServiceAsyncClient
from .services.prediction_service import PredictionServiceClient
from .services.prediction_service import PredictionServiceAsyncClient
from .services.specialist_pool_service import SpecialistPoolServiceClient
from .services.specialist_pool_service import SpecialistPoolServiceAsyncClient
from .services.tensorboard_service import TensorboardServiceClient
from .services.tensorboard_service import TensorboardServiceAsyncClient
from .services.vizier_service import VizierServiceClient
from .services.vizier_service import VizierServiceAsyncClient

from .types.accelerator_type import AcceleratorType
from .types.annotation import Annotation
from .types.annotation_spec import AnnotationSpec
from .types.artifact import Artifact
from .types.batch_prediction_job import BatchPredictionJob
from .types.completion_stats import CompletionStats
from .types.context import Context
from .types.custom_job import ContainerSpec
from .types.custom_job import CustomJob
from .types.custom_job import CustomJobSpec
from .types.custom_job import PythonPackageSpec
from .types.custom_job import Scheduling
from .types.custom_job import WorkerPoolSpec
from .types.data_item import DataItem
from .types.data_labeling_job import ActiveLearningConfig
from .types.data_labeling_job import DataLabelingJob
from .types.data_labeling_job import SampleConfig
from .types.data_labeling_job import TrainingConfig
from .types.dataset import Dataset
from .types.dataset import ExportDataConfig
from .types.dataset import ImportDataConfig
from .types.dataset_service import CreateDatasetOperationMetadata
from .types.dataset_service import CreateDatasetRequest
from .types.dataset_service import DataItemView
from .types.dataset_service import DeleteDatasetRequest
from .types.dataset_service import ExportDataOperationMetadata
from .types.dataset_service import ExportDataRequest
from .types.dataset_service import ExportDataResponse
from .types.dataset_service import GetAnnotationSpecRequest
from .types.dataset_service import GetDatasetRequest
from .types.dataset_service import ImportDataOperationMetadata
from .types.dataset_service import ImportDataRequest
from .types.dataset_service import ImportDataResponse
from .types.dataset_service import ListAnnotationsRequest
from .types.dataset_service import ListAnnotationsResponse
from .types.dataset_service import ListDataItemsRequest
from .types.dataset_service import ListDataItemsResponse
from .types.dataset_service import ListDatasetsRequest
from .types.dataset_service import ListDatasetsResponse
from .types.dataset_service import ListSavedQueriesRequest
from .types.dataset_service import ListSavedQueriesResponse
from .types.dataset_service import SearchDataItemsRequest
from .types.dataset_service import SearchDataItemsResponse
from .types.dataset_service import UpdateDatasetRequest
from .types.deployed_index_ref import DeployedIndexRef
from .types.deployed_model_ref import DeployedModelRef
from .types.deployment_resource_pool import DeploymentResourcePool
from .types.deployment_resource_pool_service import (
    CreateDeploymentResourcePoolOperationMetadata,
)
from .types.deployment_resource_pool_service import CreateDeploymentResourcePoolRequest
from .types.deployment_resource_pool_service import DeleteDeploymentResourcePoolRequest
from .types.deployment_resource_pool_service import GetDeploymentResourcePoolRequest
from .types.deployment_resource_pool_service import ListDeploymentResourcePoolsRequest
from .types.deployment_resource_pool_service import ListDeploymentResourcePoolsResponse
from .types.deployment_resource_pool_service import QueryDeployedModelsRequest
from .types.deployment_resource_pool_service import QueryDeployedModelsResponse
from .types.deployment_resource_pool_service import (
    UpdateDeploymentResourcePoolOperationMetadata,
)
from .types.encryption_spec import EncryptionSpec
from .types.endpoint import DeployedModel
from .types.endpoint import Endpoint
from .types.endpoint import PredictRequestResponseLoggingConfig
from .types.endpoint import PrivateEndpoints
from .types.endpoint_service import CreateEndpointOperationMetadata
from .types.endpoint_service import CreateEndpointRequest
from .types.endpoint_service import DeleteEndpointRequest
from .types.endpoint_service import DeployModelOperationMetadata
from .types.endpoint_service import DeployModelRequest
from .types.endpoint_service import DeployModelResponse
from .types.endpoint_service import GetEndpointRequest
from .types.endpoint_service import ListEndpointsRequest
from .types.endpoint_service import ListEndpointsResponse
from .types.endpoint_service import UndeployModelOperationMetadata
from .types.endpoint_service import UndeployModelRequest
from .types.endpoint_service import UndeployModelResponse
from .types.endpoint_service import UpdateEndpointRequest
from .types.entity_type import EntityType
from .types.env_var import EnvVar
from .types.event import Event
from .types.execution import Execution
from .types.explanation import Attribution
from .types.explanation import BlurBaselineConfig
from .types.explanation import Examples
from .types.explanation import ExamplesOverride
from .types.explanation import ExamplesRestrictionsNamespace
from .types.explanation import Explanation
from .types.explanation import ExplanationMetadataOverride
from .types.explanation import ExplanationParameters
from .types.explanation import ExplanationSpec
from .types.explanation import ExplanationSpecOverride
from .types.explanation import FeatureNoiseSigma
from .types.explanation import IntegratedGradientsAttribution
from .types.explanation import ModelExplanation
from .types.explanation import Neighbor
from .types.explanation import Presets
from .types.explanation import SampledShapleyAttribution
from .types.explanation import SmoothGradConfig
from .types.explanation import XraiAttribution
from .types.explanation_metadata import ExplanationMetadata
from .types.feature import Feature
from .types.feature_monitoring_stats import FeatureStatsAnomaly
from .types.feature_selector import FeatureSelector
from .types.feature_selector import IdMatcher
from .types.featurestore import Featurestore
from .types.featurestore_monitoring import FeaturestoreMonitoringConfig
from .types.featurestore_online_service import FeatureValue
from .types.featurestore_online_service import FeatureValueList
from .types.featurestore_online_service import ReadFeatureValuesRequest
from .types.featurestore_online_service import ReadFeatureValuesResponse
from .types.featurestore_online_service import StreamingReadFeatureValuesRequest
from .types.featurestore_online_service import WriteFeatureValuesPayload
from .types.featurestore_online_service import WriteFeatureValuesRequest
from .types.featurestore_online_service import WriteFeatureValuesResponse
from .types.featurestore_service import BatchCreateFeaturesOperationMetadata
from .types.featurestore_service import BatchCreateFeaturesRequest
from .types.featurestore_service import BatchCreateFeaturesResponse
from .types.featurestore_service import BatchReadFeatureValuesOperationMetadata
from .types.featurestore_service import BatchReadFeatureValuesRequest
from .types.featurestore_service import BatchReadFeatureValuesResponse
from .types.featurestore_service import CreateEntityTypeOperationMetadata
from .types.featurestore_service import CreateEntityTypeRequest
from .types.featurestore_service import CreateFeatureOperationMetadata
from .types.featurestore_service import CreateFeatureRequest
from .types.featurestore_service import CreateFeaturestoreOperationMetadata
from .types.featurestore_service import CreateFeaturestoreRequest
from .types.featurestore_service import DeleteEntityTypeRequest
from .types.featurestore_service import DeleteFeatureRequest
from .types.featurestore_service import DeleteFeaturestoreRequest
from .types.featurestore_service import DeleteFeatureValuesOperationMetadata
from .types.featurestore_service import DeleteFeatureValuesRequest
from .types.featurestore_service import DeleteFeatureValuesResponse
from .types.featurestore_service import DestinationFeatureSetting
from .types.featurestore_service import EntityIdSelector
from .types.featurestore_service import ExportFeatureValuesOperationMetadata
from .types.featurestore_service import ExportFeatureValuesRequest
from .types.featurestore_service import ExportFeatureValuesResponse
from .types.featurestore_service import FeatureValueDestination
from .types.featurestore_service import GetEntityTypeRequest
from .types.featurestore_service import GetFeatureRequest
from .types.featurestore_service import GetFeaturestoreRequest
from .types.featurestore_service import ImportFeatureValuesOperationMetadata
from .types.featurestore_service import ImportFeatureValuesRequest
from .types.featurestore_service import ImportFeatureValuesResponse
from .types.featurestore_service import ListEntityTypesRequest
from .types.featurestore_service import ListEntityTypesResponse
from .types.featurestore_service import ListFeaturesRequest
from .types.featurestore_service import ListFeaturesResponse
from .types.featurestore_service import ListFeaturestoresRequest
from .types.featurestore_service import ListFeaturestoresResponse
from .types.featurestore_service import SearchFeaturesRequest
from .types.featurestore_service import SearchFeaturesResponse
from .types.featurestore_service import UpdateEntityTypeRequest
from .types.featurestore_service import UpdateFeatureRequest
from .types.featurestore_service import UpdateFeaturestoreOperationMetadata
from .types.featurestore_service import UpdateFeaturestoreRequest
from .types.hyperparameter_tuning_job import HyperparameterTuningJob
from .types.index import Index
from .types.index import IndexDatapoint
from .types.index import IndexStats
from .types.index_endpoint import DeployedIndex
from .types.index_endpoint import DeployedIndexAuthConfig
from .types.index_endpoint import IndexEndpoint
from .types.index_endpoint import IndexPrivateEndpoints
from .types.index_endpoint_service import CreateIndexEndpointOperationMetadata
from .types.index_endpoint_service import CreateIndexEndpointRequest
from .types.index_endpoint_service import DeleteIndexEndpointRequest
from .types.index_endpoint_service import DeployIndexOperationMetadata
from .types.index_endpoint_service import DeployIndexRequest
from .types.index_endpoint_service import DeployIndexResponse
from .types.index_endpoint_service import GetIndexEndpointRequest
from .types.index_endpoint_service import ListIndexEndpointsRequest
from .types.index_endpoint_service import ListIndexEndpointsResponse
from .types.index_endpoint_service import MutateDeployedIndexOperationMetadata
from .types.index_endpoint_service import MutateDeployedIndexRequest
from .types.index_endpoint_service import MutateDeployedIndexResponse
from .types.index_endpoint_service import UndeployIndexOperationMetadata
from .types.index_endpoint_service import UndeployIndexRequest
from .types.index_endpoint_service import UndeployIndexResponse
from .types.index_endpoint_service import UpdateIndexEndpointRequest
from .types.index_service import CreateIndexOperationMetadata
from .types.index_service import CreateIndexRequest
from .types.index_service import DeleteIndexRequest
from .types.index_service import GetIndexRequest
from .types.index_service import ListIndexesRequest
from .types.index_service import ListIndexesResponse
from .types.index_service import NearestNeighborSearchOperationMetadata
from .types.index_service import RemoveDatapointsRequest
from .types.index_service import RemoveDatapointsResponse
from .types.index_service import UpdateIndexOperationMetadata
from .types.index_service import UpdateIndexRequest
from .types.index_service import UpsertDatapointsRequest
from .types.index_service import UpsertDatapointsResponse
from .types.io import AvroSource
from .types.io import BigQueryDestination
from .types.io import BigQuerySource
from .types.io import ContainerRegistryDestination
from .types.io import CsvDestination
from .types.io import CsvSource
from .types.io import GcsDestination
from .types.io import GcsSource
from .types.io import TFRecordDestination
from .types.job_service import CancelBatchPredictionJobRequest
from .types.job_service import CancelCustomJobRequest
from .types.job_service import CancelDataLabelingJobRequest
from .types.job_service import CancelHyperparameterTuningJobRequest
from .types.job_service import CancelNasJobRequest
from .types.job_service import CreateBatchPredictionJobRequest
from .types.job_service import CreateCustomJobRequest
from .types.job_service import CreateDataLabelingJobRequest
from .types.job_service import CreateHyperparameterTuningJobRequest
from .types.job_service import CreateModelDeploymentMonitoringJobRequest
from .types.job_service import CreateNasJobRequest
from .types.job_service import DeleteBatchPredictionJobRequest
from .types.job_service import DeleteCustomJobRequest
from .types.job_service import DeleteDataLabelingJobRequest
from .types.job_service import DeleteHyperparameterTuningJobRequest
from .types.job_service import DeleteModelDeploymentMonitoringJobRequest
from .types.job_service import DeleteNasJobRequest
from .types.job_service import GetBatchPredictionJobRequest
from .types.job_service import GetCustomJobRequest
from .types.job_service import GetDataLabelingJobRequest
from .types.job_service import GetHyperparameterTuningJobRequest
from .types.job_service import GetModelDeploymentMonitoringJobRequest
from .types.job_service import GetNasJobRequest
from .types.job_service import GetNasTrialDetailRequest
from .types.job_service import ListBatchPredictionJobsRequest
from .types.job_service import ListBatchPredictionJobsResponse
from .types.job_service import ListCustomJobsRequest
from .types.job_service import ListCustomJobsResponse
from .types.job_service import ListDataLabelingJobsRequest
from .types.job_service import ListDataLabelingJobsResponse
from .types.job_service import ListHyperparameterTuningJobsRequest
from .types.job_service import ListHyperparameterTuningJobsResponse
from .types.job_service import ListModelDeploymentMonitoringJobsRequest
from .types.job_service import ListModelDeploymentMonitoringJobsResponse
from .types.job_service import ListNasJobsRequest
from .types.job_service import ListNasJobsResponse
from .types.job_service import ListNasTrialDetailsRequest
from .types.job_service import ListNasTrialDetailsResponse
from .types.job_service import PauseModelDeploymentMonitoringJobRequest
from .types.job_service import ResumeModelDeploymentMonitoringJobRequest
from .types.job_service import SearchModelDeploymentMonitoringStatsAnomaliesRequest
from .types.job_service import SearchModelDeploymentMonitoringStatsAnomaliesResponse
from .types.job_service import UpdateModelDeploymentMonitoringJobOperationMetadata
from .types.job_service import UpdateModelDeploymentMonitoringJobRequest
from .types.job_state import JobState
from .types.lineage_subgraph import LineageSubgraph
from .types.machine_resources import AutomaticResources
from .types.machine_resources import AutoscalingMetricSpec
from .types.machine_resources import BatchDedicatedResources
from .types.machine_resources import DedicatedResources
from .types.machine_resources import DiskSpec
from .types.machine_resources import MachineSpec
from .types.machine_resources import NfsMount
from .types.machine_resources import ResourcesConsumed
from .types.manual_batch_tuning_parameters import ManualBatchTuningParameters
from .types.metadata_schema import MetadataSchema
from .types.metadata_service import AddContextArtifactsAndExecutionsRequest
from .types.metadata_service import AddContextArtifactsAndExecutionsResponse
from .types.metadata_service import AddContextChildrenRequest
from .types.metadata_service import AddContextChildrenResponse
from .types.metadata_service import AddExecutionEventsRequest
from .types.metadata_service import AddExecutionEventsResponse
from .types.metadata_service import CreateArtifactRequest
from .types.metadata_service import CreateContextRequest
from .types.metadata_service import CreateExecutionRequest
from .types.metadata_service import CreateMetadataSchemaRequest
from .types.metadata_service import CreateMetadataStoreOperationMetadata
from .types.metadata_service import CreateMetadataStoreRequest
from .types.metadata_service import DeleteArtifactRequest
from .types.metadata_service import DeleteContextRequest
from .types.metadata_service import DeleteExecutionRequest
from .types.metadata_service import DeleteMetadataStoreOperationMetadata
from .types.metadata_service import DeleteMetadataStoreRequest
from .types.metadata_service import GetArtifactRequest
from .types.metadata_service import GetContextRequest
from .types.metadata_service import GetExecutionRequest
from .types.metadata_service import GetMetadataSchemaRequest
from .types.metadata_service import GetMetadataStoreRequest
from .types.metadata_service import ListArtifactsRequest
from .types.metadata_service import ListArtifactsResponse
from .types.metadata_service import ListContextsRequest
from .types.metadata_service import ListContextsResponse
from .types.metadata_service import ListExecutionsRequest
from .types.metadata_service import ListExecutionsResponse
from .types.metadata_service import ListMetadataSchemasRequest
from .types.metadata_service import ListMetadataSchemasResponse
from .types.metadata_service import ListMetadataStoresRequest
from .types.metadata_service import ListMetadataStoresResponse
from .types.metadata_service import PurgeArtifactsMetadata
from .types.metadata_service import PurgeArtifactsRequest
from .types.metadata_service import PurgeArtifactsResponse
from .types.metadata_service import PurgeContextsMetadata
from .types.metadata_service import PurgeContextsRequest
from .types.metadata_service import PurgeContextsResponse
from .types.metadata_service import PurgeExecutionsMetadata
from .types.metadata_service import PurgeExecutionsRequest
from .types.metadata_service import PurgeExecutionsResponse
from .types.metadata_service import QueryArtifactLineageSubgraphRequest
from .types.metadata_service import QueryContextLineageSubgraphRequest
from .types.metadata_service import QueryExecutionInputsAndOutputsRequest
from .types.metadata_service import RemoveContextChildrenRequest
from .types.metadata_service import RemoveContextChildrenResponse
from .types.metadata_service import UpdateArtifactRequest
from .types.metadata_service import UpdateContextRequest
from .types.metadata_service import UpdateExecutionRequest
from .types.metadata_store import MetadataStore
from .types.migratable_resource import MigratableResource
from .types.migration_service import BatchMigrateResourcesOperationMetadata
from .types.migration_service import BatchMigrateResourcesRequest
from .types.migration_service import BatchMigrateResourcesResponse
from .types.migration_service import MigrateResourceRequest
from .types.migration_service import MigrateResourceResponse
from .types.migration_service import SearchMigratableResourcesRequest
from .types.migration_service import SearchMigratableResourcesResponse
from .types.model import Model
from .types.model import ModelContainerSpec
from .types.model import ModelSourceInfo
from .types.model import Port
from .types.model import PredictSchemata
from .types.model_deployment_monitoring_job import (
    ModelDeploymentMonitoringBigQueryTable,
)
from .types.model_deployment_monitoring_job import ModelDeploymentMonitoringJob
from .types.model_deployment_monitoring_job import (
    ModelDeploymentMonitoringObjectiveConfig,
)
from .types.model_deployment_monitoring_job import (
    ModelDeploymentMonitoringScheduleConfig,
)
from .types.model_deployment_monitoring_job import ModelMonitoringStatsAnomalies
from .types.model_deployment_monitoring_job import (
    ModelDeploymentMonitoringObjectiveType,
)
from .types.model_evaluation import ModelEvaluation
from .types.model_evaluation_slice import ModelEvaluationSlice
from .types.model_monitoring import ModelMonitoringAlertConfig
from .types.model_monitoring import ModelMonitoringConfig
from .types.model_monitoring import ModelMonitoringObjectiveConfig
from .types.model_monitoring import SamplingStrategy
from .types.model_monitoring import ThresholdConfig
from .types.model_service import BatchImportModelEvaluationSlicesRequest
from .types.model_service import BatchImportModelEvaluationSlicesResponse
from .types.model_service import CopyModelOperationMetadata
from .types.model_service import CopyModelRequest
from .types.model_service import CopyModelResponse
from .types.model_service import DeleteModelRequest
from .types.model_service import DeleteModelVersionRequest
from .types.model_service import ExportModelOperationMetadata
from .types.model_service import ExportModelRequest
from .types.model_service import ExportModelResponse
from .types.model_service import GetModelEvaluationRequest
from .types.model_service import GetModelEvaluationSliceRequest
from .types.model_service import GetModelRequest
from .types.model_service import ImportModelEvaluationRequest
from .types.model_service import ListModelEvaluationSlicesRequest
from .types.model_service import ListModelEvaluationSlicesResponse
from .types.model_service import ListModelEvaluationsRequest
from .types.model_service import ListModelEvaluationsResponse
from .types.model_service import ListModelsRequest
from .types.model_service import ListModelsResponse
from .types.model_service import ListModelVersionsRequest
from .types.model_service import ListModelVersionsResponse
from .types.model_service import MergeVersionAliasesRequest
from .types.model_service import UpdateExplanationDatasetOperationMetadata
from .types.model_service import UpdateExplanationDatasetRequest
from .types.model_service import UpdateExplanationDatasetResponse
from .types.model_service import UpdateModelRequest
from .types.model_service import UploadModelOperationMetadata
from .types.model_service import UploadModelRequest
from .types.model_service import UploadModelResponse
from .types.nas_job import NasJob
from .types.nas_job import NasJobOutput
from .types.nas_job import NasJobSpec
from .types.nas_job import NasTrial
from .types.nas_job import NasTrialDetail
from .types.operation import DeleteOperationMetadata
from .types.operation import GenericOperationMetadata
from .types.pipeline_failure_policy import PipelineFailurePolicy
from .types.pipeline_job import PipelineJob
from .types.pipeline_job import PipelineJobDetail
from .types.pipeline_job import PipelineTaskDetail
from .types.pipeline_job import PipelineTaskExecutorDetail
from .types.pipeline_job import PipelineTemplateMetadata
from .types.pipeline_service import CancelPipelineJobRequest
from .types.pipeline_service import CancelTrainingPipelineRequest
from .types.pipeline_service import CreatePipelineJobRequest
from .types.pipeline_service import CreateTrainingPipelineRequest
from .types.pipeline_service import DeletePipelineJobRequest
from .types.pipeline_service import DeleteTrainingPipelineRequest
from .types.pipeline_service import GetPipelineJobRequest
from .types.pipeline_service import GetTrainingPipelineRequest
from .types.pipeline_service import ListPipelineJobsRequest
from .types.pipeline_service import ListPipelineJobsResponse
from .types.pipeline_service import ListTrainingPipelinesRequest
from .types.pipeline_service import ListTrainingPipelinesResponse
from .types.pipeline_state import PipelineState
from .types.prediction_service import ExplainRequest
from .types.prediction_service import ExplainResponse
from .types.prediction_service import PredictRequest
from .types.prediction_service import PredictResponse
from .types.prediction_service import RawPredictRequest
from .types.saved_query import SavedQuery
from .types.service_networking import PrivateServiceConnectConfig
from .types.specialist_pool import SpecialistPool
from .types.specialist_pool_service import CreateSpecialistPoolOperationMetadata
from .types.specialist_pool_service import CreateSpecialistPoolRequest
from .types.specialist_pool_service import DeleteSpecialistPoolRequest
from .types.specialist_pool_service import GetSpecialistPoolRequest
from .types.specialist_pool_service import ListSpecialistPoolsRequest
from .types.specialist_pool_service import ListSpecialistPoolsResponse
from .types.specialist_pool_service import UpdateSpecialistPoolOperationMetadata
from .types.specialist_pool_service import UpdateSpecialistPoolRequest
from .types.study import Measurement
from .types.study import Study
from .types.study import StudySpec
from .types.study import Trial
from .types.tensorboard import Tensorboard
from .types.tensorboard_data import Scalar
from .types.tensorboard_data import TensorboardBlob
from .types.tensorboard_data import TensorboardBlobSequence
from .types.tensorboard_data import TensorboardTensor
from .types.tensorboard_data import TimeSeriesData
from .types.tensorboard_data import TimeSeriesDataPoint
from .types.tensorboard_experiment import TensorboardExperiment
from .types.tensorboard_run import TensorboardRun
from .types.tensorboard_service import BatchCreateTensorboardRunsRequest
from .types.tensorboard_service import BatchCreateTensorboardRunsResponse
from .types.tensorboard_service import BatchCreateTensorboardTimeSeriesRequest
from .types.tensorboard_service import BatchCreateTensorboardTimeSeriesResponse
from .types.tensorboard_service import BatchReadTensorboardTimeSeriesDataRequest
from .types.tensorboard_service import BatchReadTensorboardTimeSeriesDataResponse
from .types.tensorboard_service import CreateTensorboardExperimentRequest
from .types.tensorboard_service import CreateTensorboardOperationMetadata
from .types.tensorboard_service import CreateTensorboardRequest
from .types.tensorboard_service import CreateTensorboardRunRequest
from .types.tensorboard_service import CreateTensorboardTimeSeriesRequest
from .types.tensorboard_service import DeleteTensorboardExperimentRequest
from .types.tensorboard_service import DeleteTensorboardRequest
from .types.tensorboard_service import DeleteTensorboardRunRequest
from .types.tensorboard_service import DeleteTensorboardTimeSeriesRequest
from .types.tensorboard_service import ExportTensorboardTimeSeriesDataRequest
from .types.tensorboard_service import ExportTensorboardTimeSeriesDataResponse
from .types.tensorboard_service import GetTensorboardExperimentRequest
from .types.tensorboard_service import GetTensorboardRequest
from .types.tensorboard_service import GetTensorboardRunRequest
from .types.tensorboard_service import GetTensorboardTimeSeriesRequest
from .types.tensorboard_service import ListTensorboardExperimentsRequest
from .types.tensorboard_service import ListTensorboardExperimentsResponse
from .types.tensorboard_service import ListTensorboardRunsRequest
from .types.tensorboard_service import ListTensorboardRunsResponse
from .types.tensorboard_service import ListTensorboardsRequest
from .types.tensorboard_service import ListTensorboardsResponse
from .types.tensorboard_service import ListTensorboardTimeSeriesRequest
from .types.tensorboard_service import ListTensorboardTimeSeriesResponse
from .types.tensorboard_service import ReadTensorboardBlobDataRequest
from .types.tensorboard_service import ReadTensorboardBlobDataResponse
from .types.tensorboard_service import ReadTensorboardTimeSeriesDataRequest
from .types.tensorboard_service import ReadTensorboardTimeSeriesDataResponse
from .types.tensorboard_service import ReadTensorboardUsageRequest
from .types.tensorboard_service import ReadTensorboardUsageResponse
from .types.tensorboard_service import UpdateTensorboardExperimentRequest
from .types.tensorboard_service import UpdateTensorboardOperationMetadata
from .types.tensorboard_service import UpdateTensorboardRequest
from .types.tensorboard_service import UpdateTensorboardRunRequest
from .types.tensorboard_service import UpdateTensorboardTimeSeriesRequest
from .types.tensorboard_service import WriteTensorboardExperimentDataRequest
from .types.tensorboard_service import WriteTensorboardExperimentDataResponse
from .types.tensorboard_service import WriteTensorboardRunDataRequest
from .types.tensorboard_service import WriteTensorboardRunDataResponse
from .types.tensorboard_time_series import TensorboardTimeSeries
from .types.training_pipeline import FilterSplit
from .types.training_pipeline import FractionSplit
from .types.training_pipeline import InputDataConfig
from .types.training_pipeline import PredefinedSplit
from .types.training_pipeline import StratifiedSplit
from .types.training_pipeline import TimestampSplit
from .types.training_pipeline import TrainingPipeline
from .types.types import BoolArray
from .types.types import DoubleArray
from .types.types import Int64Array
from .types.types import StringArray
from .types.unmanaged_container_model import UnmanagedContainerModel
from .types.user_action_reference import UserActionReference
from .types.value import Value
from .types.vizier_service import AddTrialMeasurementRequest
from .types.vizier_service import CheckTrialEarlyStoppingStateMetatdata
from .types.vizier_service import CheckTrialEarlyStoppingStateRequest
from .types.vizier_service import CheckTrialEarlyStoppingStateResponse
from .types.vizier_service import CompleteTrialRequest
from .types.vizier_service import CreateStudyRequest
from .types.vizier_service import CreateTrialRequest
from .types.vizier_service import DeleteStudyRequest
from .types.vizier_service import DeleteTrialRequest
from .types.vizier_service import GetStudyRequest
from .types.vizier_service import GetTrialRequest
from .types.vizier_service import ListOptimalTrialsRequest
from .types.vizier_service import ListOptimalTrialsResponse
from .types.vizier_service import ListStudiesRequest
from .types.vizier_service import ListStudiesResponse
from .types.vizier_service import ListTrialsRequest
from .types.vizier_service import ListTrialsResponse
from .types.vizier_service import LookupStudyRequest
from .types.vizier_service import StopTrialRequest
from .types.vizier_service import SuggestTrialsMetadata
from .types.vizier_service import SuggestTrialsRequest
from .types.vizier_service import SuggestTrialsResponse

__all__ = (
    "DatasetServiceAsyncClient",
    "DeploymentResourcePoolServiceAsyncClient",
    "EndpointServiceAsyncClient",
    "FeaturestoreOnlineServingServiceAsyncClient",
    "FeaturestoreServiceAsyncClient",
    "IndexEndpointServiceAsyncClient",
    "IndexServiceAsyncClient",
    "JobServiceAsyncClient",
    "MetadataServiceAsyncClient",
    "MigrationServiceAsyncClient",
    "ModelServiceAsyncClient",
    "PipelineServiceAsyncClient",
    "PredictionServiceAsyncClient",
    "SpecialistPoolServiceAsyncClient",
    "TensorboardServiceAsyncClient",
    "VizierServiceAsyncClient",
    "AcceleratorType",
    "ActiveLearningConfig",
    "AddContextArtifactsAndExecutionsRequest",
    "AddContextArtifactsAndExecutionsResponse",
    "AddContextChildrenRequest",
    "AddContextChildrenResponse",
    "AddExecutionEventsRequest",
    "AddExecutionEventsResponse",
    "AddTrialMeasurementRequest",
    "Annotation",
    "AnnotationSpec",
    "Artifact",
    "Attribution",
    "AutomaticResources",
    "AutoscalingMetricSpec",
    "AvroSource",
    "BatchCreateFeaturesOperationMetadata",
    "BatchCreateFeaturesRequest",
    "BatchCreateFeaturesResponse",
    "BatchCreateTensorboardRunsRequest",
    "BatchCreateTensorboardRunsResponse",
    "BatchCreateTensorboardTimeSeriesRequest",
    "BatchCreateTensorboardTimeSeriesResponse",
    "BatchDedicatedResources",
    "BatchImportModelEvaluationSlicesRequest",
    "BatchImportModelEvaluationSlicesResponse",
    "BatchMigrateResourcesOperationMetadata",
    "BatchMigrateResourcesRequest",
    "BatchMigrateResourcesResponse",
    "BatchPredictionJob",
    "BatchReadFeatureValuesOperationMetadata",
    "BatchReadFeatureValuesRequest",
    "BatchReadFeatureValuesResponse",
    "BatchReadTensorboardTimeSeriesDataRequest",
    "BatchReadTensorboardTimeSeriesDataResponse",
    "BigQueryDestination",
    "BigQuerySource",
    "BlurBaselineConfig",
    "BoolArray",
    "CancelBatchPredictionJobRequest",
    "CancelCustomJobRequest",
    "CancelDataLabelingJobRequest",
    "CancelHyperparameterTuningJobRequest",
    "CancelNasJobRequest",
    "CancelPipelineJobRequest",
    "CancelTrainingPipelineRequest",
    "CheckTrialEarlyStoppingStateMetatdata",
    "CheckTrialEarlyStoppingStateRequest",
    "CheckTrialEarlyStoppingStateResponse",
    "CompleteTrialRequest",
    "CompletionStats",
    "ContainerRegistryDestination",
    "ContainerSpec",
    "Context",
    "CopyModelOperationMetadata",
    "CopyModelRequest",
    "CopyModelResponse",
    "CreateArtifactRequest",
    "CreateBatchPredictionJobRequest",
    "CreateContextRequest",
    "CreateCustomJobRequest",
    "CreateDataLabelingJobRequest",
    "CreateDatasetOperationMetadata",
    "CreateDatasetRequest",
    "CreateDeploymentResourcePoolOperationMetadata",
    "CreateDeploymentResourcePoolRequest",
    "CreateEndpointOperationMetadata",
    "CreateEndpointRequest",
    "CreateEntityTypeOperationMetadata",
    "CreateEntityTypeRequest",
    "CreateExecutionRequest",
    "CreateFeatureOperationMetadata",
    "CreateFeatureRequest",
    "CreateFeaturestoreOperationMetadata",
    "CreateFeaturestoreRequest",
    "CreateHyperparameterTuningJobRequest",
    "CreateIndexEndpointOperationMetadata",
    "CreateIndexEndpointRequest",
    "CreateIndexOperationMetadata",
    "CreateIndexRequest",
    "CreateMetadataSchemaRequest",
    "CreateMetadataStoreOperationMetadata",
    "CreateMetadataStoreRequest",
    "CreateModelDeploymentMonitoringJobRequest",
    "CreateNasJobRequest",
    "CreatePipelineJobRequest",
    "CreateSpecialistPoolOperationMetadata",
    "CreateSpecialistPoolRequest",
    "CreateStudyRequest",
    "CreateTensorboardExperimentRequest",
    "CreateTensorboardOperationMetadata",
    "CreateTensorboardRequest",
    "CreateTensorboardRunRequest",
    "CreateTensorboardTimeSeriesRequest",
    "CreateTrainingPipelineRequest",
    "CreateTrialRequest",
    "CsvDestination",
    "CsvSource",
    "CustomJob",
    "CustomJobSpec",
    "DataItem",
    "DataItemView",
    "DataLabelingJob",
    "Dataset",
    "DatasetServiceClient",
    "DedicatedResources",
    "DeleteArtifactRequest",
    "DeleteBatchPredictionJobRequest",
    "DeleteContextRequest",
    "DeleteCustomJobRequest",
    "DeleteDataLabelingJobRequest",
    "DeleteDatasetRequest",
    "DeleteDeploymentResourcePoolRequest",
    "DeleteEndpointRequest",
    "DeleteEntityTypeRequest",
    "DeleteExecutionRequest",
    "DeleteFeatureRequest",
    "DeleteFeatureValuesOperationMetadata",
    "DeleteFeatureValuesRequest",
    "DeleteFeatureValuesResponse",
    "DeleteFeaturestoreRequest",
    "DeleteHyperparameterTuningJobRequest",
    "DeleteIndexEndpointRequest",
    "DeleteIndexRequest",
    "DeleteMetadataStoreOperationMetadata",
    "DeleteMetadataStoreRequest",
    "DeleteModelDeploymentMonitoringJobRequest",
    "DeleteModelRequest",
    "DeleteModelVersionRequest",
    "DeleteNasJobRequest",
    "DeleteOperationMetadata",
    "DeletePipelineJobRequest",
    "DeleteSpecialistPoolRequest",
    "DeleteStudyRequest",
    "DeleteTensorboardExperimentRequest",
    "DeleteTensorboardRequest",
    "DeleteTensorboardRunRequest",
    "DeleteTensorboardTimeSeriesRequest",
    "DeleteTrainingPipelineRequest",
    "DeleteTrialRequest",
    "DeployIndexOperationMetadata",
    "DeployIndexRequest",
    "DeployIndexResponse",
    "DeployModelOperationMetadata",
    "DeployModelRequest",
    "DeployModelResponse",
    "DeployedIndex",
    "DeployedIndexAuthConfig",
    "DeployedIndexRef",
    "DeployedModel",
    "DeployedModelRef",
    "DeploymentResourcePool",
    "DeploymentResourcePoolServiceClient",
    "DestinationFeatureSetting",
    "DiskSpec",
    "DoubleArray",
    "EncryptionSpec",
    "Endpoint",
    "EndpointServiceClient",
    "EntityIdSelector",
    "EntityType",
    "EnvVar",
    "Event",
    "Examples",
    "ExamplesOverride",
    "ExamplesRestrictionsNamespace",
    "Execution",
    "ExplainRequest",
    "ExplainResponse",
    "Explanation",
    "ExplanationMetadata",
    "ExplanationMetadataOverride",
    "ExplanationParameters",
    "ExplanationSpec",
    "ExplanationSpecOverride",
    "ExportDataConfig",
    "ExportDataOperationMetadata",
    "ExportDataRequest",
    "ExportDataResponse",
    "ExportFeatureValuesOperationMetadata",
    "ExportFeatureValuesRequest",
    "ExportFeatureValuesResponse",
    "ExportModelOperationMetadata",
    "ExportModelRequest",
    "ExportModelResponse",
    "ExportTensorboardTimeSeriesDataRequest",
    "ExportTensorboardTimeSeriesDataResponse",
    "Feature",
    "FeatureNoiseSigma",
    "FeatureSelector",
    "FeatureStatsAnomaly",
    "FeatureValue",
    "FeatureValueDestination",
    "FeatureValueList",
    "Featurestore",
    "FeaturestoreMonitoringConfig",
    "FeaturestoreOnlineServingServiceClient",
    "FeaturestoreServiceClient",
    "FilterSplit",
    "FractionSplit",
    "GcsDestination",
    "GcsSource",
    "GenericOperationMetadata",
    "GetAnnotationSpecRequest",
    "GetArtifactRequest",
    "GetBatchPredictionJobRequest",
    "GetContextRequest",
    "GetCustomJobRequest",
    "GetDataLabelingJobRequest",
    "GetDatasetRequest",
    "GetDeploymentResourcePoolRequest",
    "GetEndpointRequest",
    "GetEntityTypeRequest",
    "GetExecutionRequest",
    "GetFeatureRequest",
    "GetFeaturestoreRequest",
    "GetHyperparameterTuningJobRequest",
    "GetIndexEndpointRequest",
    "GetIndexRequest",
    "GetMetadataSchemaRequest",
    "GetMetadataStoreRequest",
    "GetModelDeploymentMonitoringJobRequest",
    "GetModelEvaluationRequest",
    "GetModelEvaluationSliceRequest",
    "GetModelRequest",
    "GetNasJobRequest",
    "GetNasTrialDetailRequest",
    "GetPipelineJobRequest",
    "GetSpecialistPoolRequest",
    "GetStudyRequest",
    "GetTensorboardExperimentRequest",
    "GetTensorboardRequest",
    "GetTensorboardRunRequest",
    "GetTensorboardTimeSeriesRequest",
    "GetTrainingPipelineRequest",
    "GetTrialRequest",
    "HyperparameterTuningJob",
    "IdMatcher",
    "ImportDataConfig",
    "ImportDataOperationMetadata",
    "ImportDataRequest",
    "ImportDataResponse",
    "ImportFeatureValuesOperationMetadata",
    "ImportFeatureValuesRequest",
    "ImportFeatureValuesResponse",
    "ImportModelEvaluationRequest",
    "Index",
    "IndexDatapoint",
    "IndexEndpoint",
    "IndexEndpointServiceClient",
    "IndexPrivateEndpoints",
    "IndexServiceClient",
    "IndexStats",
    "InputDataConfig",
    "Int64Array",
    "IntegratedGradientsAttribution",
    "JobServiceClient",
    "JobState",
    "LineageSubgraph",
    "ListAnnotationsRequest",
    "ListAnnotationsResponse",
    "ListArtifactsRequest",
    "ListArtifactsResponse",
    "ListBatchPredictionJobsRequest",
    "ListBatchPredictionJobsResponse",
    "ListContextsRequest",
    "ListContextsResponse",
    "ListCustomJobsRequest",
    "ListCustomJobsResponse",
    "ListDataItemsRequest",
    "ListDataItemsResponse",
    "ListDataLabelingJobsRequest",
    "ListDataLabelingJobsResponse",
    "ListDatasetsRequest",
    "ListDatasetsResponse",
    "ListDeploymentResourcePoolsRequest",
    "ListDeploymentResourcePoolsResponse",
    "ListEndpointsRequest",
    "ListEndpointsResponse",
    "ListEntityTypesRequest",
    "ListEntityTypesResponse",
    "ListExecutionsRequest",
    "ListExecutionsResponse",
    "ListFeaturesRequest",
    "ListFeaturesResponse",
    "ListFeaturestoresRequest",
    "ListFeaturestoresResponse",
    "ListHyperparameterTuningJobsRequest",
    "ListHyperparameterTuningJobsResponse",
    "ListIndexEndpointsRequest",
    "ListIndexEndpointsResponse",
    "ListIndexesRequest",
    "ListIndexesResponse",
    "ListMetadataSchemasRequest",
    "ListMetadataSchemasResponse",
    "ListMetadataStoresRequest",
    "ListMetadataStoresResponse",
    "ListModelDeploymentMonitoringJobsRequest",
    "ListModelDeploymentMonitoringJobsResponse",
    "ListModelEvaluationSlicesRequest",
    "ListModelEvaluationSlicesResponse",
    "ListModelEvaluationsRequest",
    "ListModelEvaluationsResponse",
    "ListModelVersionsRequest",
    "ListModelVersionsResponse",
    "ListModelsRequest",
    "ListModelsResponse",
    "ListNasJobsRequest",
    "ListNasJobsResponse",
    "ListNasTrialDetailsRequest",
    "ListNasTrialDetailsResponse",
    "ListOptimalTrialsRequest",
    "ListOptimalTrialsResponse",
    "ListPipelineJobsRequest",
    "ListPipelineJobsResponse",
    "ListSavedQueriesRequest",
    "ListSavedQueriesResponse",
    "ListSpecialistPoolsRequest",
    "ListSpecialistPoolsResponse",
    "ListStudiesRequest",
    "ListStudiesResponse",
    "ListTensorboardExperimentsRequest",
    "ListTensorboardExperimentsResponse",
    "ListTensorboardRunsRequest",
    "ListTensorboardRunsResponse",
    "ListTensorboardTimeSeriesRequest",
    "ListTensorboardTimeSeriesResponse",
    "ListTensorboardsRequest",
    "ListTensorboardsResponse",
    "ListTrainingPipelinesRequest",
    "ListTrainingPipelinesResponse",
    "ListTrialsRequest",
    "ListTrialsResponse",
    "LookupStudyRequest",
    "MachineSpec",
    "ManualBatchTuningParameters",
    "Measurement",
    "MergeVersionAliasesRequest",
    "MetadataSchema",
    "MetadataServiceClient",
    "MetadataStore",
    "MigratableResource",
    "MigrateResourceRequest",
    "MigrateResourceResponse",
    "MigrationServiceClient",
    "Model",
    "ModelContainerSpec",
    "ModelDeploymentMonitoringBigQueryTable",
    "ModelDeploymentMonitoringJob",
    "ModelDeploymentMonitoringObjectiveConfig",
    "ModelDeploymentMonitoringObjectiveType",
    "ModelDeploymentMonitoringScheduleConfig",
    "ModelEvaluation",
    "ModelEvaluationSlice",
    "ModelExplanation",
    "ModelMonitoringAlertConfig",
    "ModelMonitoringConfig",
    "ModelMonitoringObjectiveConfig",
    "ModelMonitoringStatsAnomalies",
    "ModelServiceClient",
    "ModelSourceInfo",
    "MutateDeployedIndexOperationMetadata",
    "MutateDeployedIndexRequest",
    "MutateDeployedIndexResponse",
    "NasJob",
    "NasJobOutput",
    "NasJobSpec",
    "NasTrial",
    "NasTrialDetail",
    "NearestNeighborSearchOperationMetadata",
    "Neighbor",
    "NfsMount",
    "PauseModelDeploymentMonitoringJobRequest",
    "PipelineFailurePolicy",
    "PipelineJob",
    "PipelineJobDetail",
    "PipelineServiceClient",
    "PipelineState",
    "PipelineTaskDetail",
    "PipelineTaskExecutorDetail",
    "PipelineTemplateMetadata",
    "Port",
    "PredefinedSplit",
    "PredictRequest",
    "PredictRequestResponseLoggingConfig",
    "PredictResponse",
    "PredictSchemata",
    "PredictionServiceClient",
    "Presets",
    "PrivateEndpoints",
    "PrivateServiceConnectConfig",
    "PurgeArtifactsMetadata",
    "PurgeArtifactsRequest",
    "PurgeArtifactsResponse",
    "PurgeContextsMetadata",
    "PurgeContextsRequest",
    "PurgeContextsResponse",
    "PurgeExecutionsMetadata",
    "PurgeExecutionsRequest",
    "PurgeExecutionsResponse",
    "PythonPackageSpec",
    "QueryArtifactLineageSubgraphRequest",
    "QueryContextLineageSubgraphRequest",
    "QueryDeployedModelsRequest",
    "QueryDeployedModelsResponse",
    "QueryExecutionInputsAndOutputsRequest",
    "RawPredictRequest",
    "ReadFeatureValuesRequest",
    "ReadFeatureValuesResponse",
    "ReadTensorboardBlobDataRequest",
    "ReadTensorboardBlobDataResponse",
    "ReadTensorboardTimeSeriesDataRequest",
    "ReadTensorboardTimeSeriesDataResponse",
    "ReadTensorboardUsageRequest",
    "ReadTensorboardUsageResponse",
    "RemoveContextChildrenRequest",
    "RemoveContextChildrenResponse",
    "RemoveDatapointsRequest",
    "RemoveDatapointsResponse",
    "ResourcesConsumed",
    "ResumeModelDeploymentMonitoringJobRequest",
    "SampleConfig",
    "SampledShapleyAttribution",
    "SamplingStrategy",
    "SavedQuery",
    "Scalar",
    "Scheduling",
    "SearchDataItemsRequest",
    "SearchDataItemsResponse",
    "SearchFeaturesRequest",
    "SearchFeaturesResponse",
    "SearchMigratableResourcesRequest",
    "SearchMigratableResourcesResponse",
    "SearchModelDeploymentMonitoringStatsAnomaliesRequest",
    "SearchModelDeploymentMonitoringStatsAnomaliesResponse",
    "SmoothGradConfig",
    "SpecialistPool",
    "SpecialistPoolServiceClient",
    "StopTrialRequest",
    "StratifiedSplit",
    "StreamingReadFeatureValuesRequest",
    "StringArray",
    "Study",
    "StudySpec",
    "SuggestTrialsMetadata",
    "SuggestTrialsRequest",
    "SuggestTrialsResponse",
    "TFRecordDestination",
    "Tensorboard",
    "TensorboardBlob",
    "TensorboardBlobSequence",
    "TensorboardExperiment",
    "TensorboardRun",
    "TensorboardServiceClient",
    "TensorboardTensor",
    "TensorboardTimeSeries",
    "ThresholdConfig",
    "TimeSeriesData",
    "TimeSeriesDataPoint",
    "TimestampSplit",
    "TrainingConfig",
    "TrainingPipeline",
    "Trial",
    "UndeployIndexOperationMetadata",
    "UndeployIndexRequest",
    "UndeployIndexResponse",
    "UndeployModelOperationMetadata",
    "UndeployModelRequest",
    "UndeployModelResponse",
    "UnmanagedContainerModel",
    "UpdateArtifactRequest",
    "UpdateContextRequest",
    "UpdateDatasetRequest",
    "UpdateDeploymentResourcePoolOperationMetadata",
    "UpdateEndpointRequest",
    "UpdateEntityTypeRequest",
    "UpdateExecutionRequest",
    "UpdateExplanationDatasetOperationMetadata",
    "UpdateExplanationDatasetRequest",
    "UpdateExplanationDatasetResponse",
    "UpdateFeatureRequest",
    "UpdateFeaturestoreOperationMetadata",
    "UpdateFeaturestoreRequest",
    "UpdateIndexEndpointRequest",
    "UpdateIndexOperationMetadata",
    "UpdateIndexRequest",
    "UpdateModelDeploymentMonitoringJobOperationMetadata",
    "UpdateModelDeploymentMonitoringJobRequest",
    "UpdateModelRequest",
    "UpdateSpecialistPoolOperationMetadata",
    "UpdateSpecialistPoolRequest",
    "UpdateTensorboardExperimentRequest",
    "UpdateTensorboardOperationMetadata",
    "UpdateTensorboardRequest",
    "UpdateTensorboardRunRequest",
    "UpdateTensorboardTimeSeriesRequest",
    "UploadModelOperationMetadata",
    "UploadModelRequest",
    "UploadModelResponse",
    "UpsertDatapointsRequest",
    "UpsertDatapointsResponse",
    "UserActionReference",
    "Value",
    "VizierServiceClient",
    "WorkerPoolSpec",
    "WriteFeatureValuesPayload",
    "WriteFeatureValuesRequest",
    "WriteFeatureValuesResponse",
    "WriteTensorboardExperimentDataRequest",
    "WriteTensorboardExperimentDataResponse",
    "WriteTensorboardRunDataRequest",
    "WriteTensorboardRunDataResponse",
    "XraiAttribution",
)



# -*- coding: utf-8 -*-

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from importlib import reload
import pytest
import time
from typing import Optional

from google.cloud.aiplatform import base
from google.cloud.aiplatform import initializer


class _TestClass(base.FutureManager):
    def __init__(self, x):
        self.x = x
        super().__init__()

    @classmethod
    def _empty_constructor(cls):
        self = cls.__new__(cls)
        base.FutureManager.__init__(self)
        self.x = None
        return self

    def _sync_object_with_future_result(self, result):
        self.x = result.x

    @classmethod
    @base.optional_sync()
    def create(cls, x: int, sync=True) -> "_TestClass":
        time.sleep(1)
        return cls(x)

    @base.optional_sync()
    def add(self, a: "_TestClass", sync=True) -> None:
        time.sleep(1)
        return self._add(a=a, sync=sync)

    def _add(self, a: "_TestClass", sync=True) -> None:
        self.x = self.x + a.x


class _TestClassDownStream(_TestClass):
    @base.optional_sync(construct_object_on_arg="a")
    def add_and_create_new(
        self, a: Optional["_TestClass"] = None, sync=True
    ) -> _TestClass:
        time.sleep(1)
        if a:
            return _TestClass(self.x + a.x)
        return None

    @base.optional_sync(return_input_arg="a", bind_future_to_self=False)
    def add_to_input_arg(self, a: "_TestClass", sync=True) -> _TestClass:
        time.sleep(1)
        a._add(self)
        return a


class TestFutureManager:
    def setup_method(self):
        reload(initializer)

    def teardown_method(self):
        initializer.global_pool.shutdown(wait=True)

    @pytest.mark.parametrize("sync", [True, False])
    def test_create_task(self, sync):
        a = _TestClass.create(10, sync=sync)
        if not sync:
            assert a.x is None
            assert a._latest_future is not None
            a.wait()
        assert a._latest_future is None
        assert a.x == 10
        assert isinstance(a, _TestClass)

    @pytest.mark.parametrize("sync", [True, False])
    def test_create_and_add_task(self, sync):
        _latest_future = None

        a = _TestClass.create(10, sync=sync)
        b = _TestClass.create(7, sync=sync)
        if not sync:
            assert a.x is None
            assert a._latest_future is not None
            assert b.x is None
            assert b._latest_future is not None
            _latest_future = b._latest_future

        b.add(a, sync=sync)

        if not sync:
            assert b._latest_future is not _latest_future
            b.wait()

        assert a._latest_future is None
        assert a.x == 10
        assert b._latest_future is None
        assert b.x == 17
        assert isinstance(a, _TestClass)
        assert isinstance(b, _TestClass)

    @pytest.mark.parametrize("sync", [True, False])
    def test_create_and_add_and_create_new_task(self, sync):
        _latest_future = None

        a = _TestClass.create(10, sync=sync)
        b = _TestClassDownStream.create(7, sync=sync)
        if not sync:
            assert a.x is None
            assert a._latest_future is not None
            assert b.x is None
            assert b._latest_future is not None
            _latest_future = b._latest_future

        c = b.add_and_create_new(a, sync=sync)

        if not sync:
            assert b._latest_future is not _latest_future
            assert c.x is None
            assert c._latest_future is not None
            c.wait()

        assert a._latest_future is None
        assert a.x == 10
        assert b._latest_future is None
        assert b.x == 7
        assert c._latest_future is None
        assert c.x == 17
        assert isinstance(a, _TestClass)
        assert isinstance(b, _TestClassDownStream)
        assert isinstance(c, _TestClass)

    @pytest.mark.parametrize("sync", [True, False])
    def test_create_and_add_and_not_create_new_task(self, sync):
        _latest_future = None

        b = _TestClassDownStream.create(7, sync=sync)
        if not sync:
            assert b.x is None
            assert b._latest_future is not None
            _latest_future = b._latest_future

        c = b.add_and_create_new(None, sync=sync)

        if not sync:
            assert b._latest_future is not _latest_future
            b.wait()

        assert c is None

        assert b._latest_future is None
        assert b.x == 7
        assert isinstance(b, _TestClassDownStream)

    @pytest.mark.parametrize("sync", [True, False])
    def test_create_and_add_return_arg(self, sync):
        _latest_future = None

        a = _TestClass.create(10, sync=sync)
        b = _TestClassDownStream.create(7, sync=sync)
        if not sync:
            assert a.x is None
            assert a._latest_future is not None
            assert b.x is None
            assert b._latest_future is not None
            _latest_future = b._latest_future

        c = b.add_to_input_arg(a, sync=sync)

        if not sync:
            assert b._latest_future is _latest_future
            assert c.x is None
            assert c._latest_future is not None
            assert c is a
            c.wait()

        assert a._latest_future is None
        assert a.x == 17
        assert b._latest_future is None
        assert b.x == 7
        assert c._latest_future is None
        assert c.x == 17
        assert isinstance(a, _TestClass)
        assert isinstance(b, _TestClassDownStream)
        assert isinstance(c, _TestClass)
